/*
 * Copyright (C) 1995 Lars Fenneberg
 * Portions Copyright (C) 2000 Erik Cumps
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* 
 * New implementation of hash.c (ipquota in-memory database access API)
 * using MIDBS
 *
 * $Id: hash.c,v 1.22 2000/10/12 20:57:52 slay Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/ip_acct_user.h>
#include "ipquota.h"
#include <midbs.h>

#ifndef lint
static const char _vcid[] = "$Id: hash.c,v 1.22 2000/10/12 20:57:52 slay Exp $";
#endif /* lint */

void init_hash(void)
{
    return;
}

int exist_user(uid_t uid)
{
    search_handle sh;
    struct search_info si;
    struct user_info ui;
    int rc;

    if (MIDBS_allocsearch (&sh) == -1) {
	mylog ("exist_user: error: no backend module linked in!");
	exit (EXIT_FAILURE);
    }

    si.uid_first = uid;
    si.uid_last  = uid;
    rc = MIDBS_findfirst (sh, &si, &ui);
    MIDBS_freesearch (&sh);

    return (rc == 0);
}

void add_user(uid_t uid, signed long long Hquota, signed long long Squota, time_t maxT, time_t T0, int flags, int warn)
{
    struct user_info ui;
    int rc;

    ui.uid    = uid;
    ui.Hquota = Hquota;
    ui.Squota = Squota;
    ui.maxT   = maxT;
    ui.T0     = T0;
    ui.flags  = flags;
    ui.warn   = warn;
    rc = MIDBS_write (&ui);

    if (rc == -1) {
	mylog ("add_user: error: no backend module linked in!");
	exit (EXIT_FAILURE);
    }
    if (rc == -2) {
	mylog ("add_user: error: failed to add record");
	exit (EXIT_FAILURE);
    }
    if (rc == 0) {
	mylog ("add_user: warning: record was updated");
    }

    return;
}

void remove_user(uid_t uid)
{
    struct user_info p;
    int rc;

    p.uid = uid;
    rc = MIDBS_delete (&p);

    if (rc == -1) {
	mylog ("remove_user: error: no backend module linked in!");
	exit (EXIT_FAILURE);
    }
    if (rc == -2) {
	mylog ("remove_user: warning: uid %lu not found in quota database", uid);
    } else {
        /* enable ip access for uid */
	MIDBS_ctrl_access (uid, ~IPAU_NO_ACCESS);
        /*** FIXME - should check returncode and deal with it */
    }
}

void ctrl_access(uid_t uid, int flags)
{
    MIDBS_ctrl_access (uid, flags);
    /*** FIXME - should check returncode and deal with it */
}

void ctrl_access_all(void)
{
    search_handle sh;
    struct search_info si;
    struct user_info ui;
    int rc;

    if (MIDBS_allocsearch (&sh) == -1) {
	mylog ("ctrl_access_all: error: no backend module linked in!");
	exit (EXIT_FAILURE);
    }

    si.uid_first = 0;
    si.uid_last  = 1;
    rc = MIDBS_findfirst (sh, &si, &ui);

    if (rc == -2) {
	MIDBS_freesearch (&sh);
	return;
    }

    ctrl_access (ui.uid, ui.flags);
    while (MIDBS_findnext (sh, &ui) == 0) {
	ctrl_access (ui.uid, ui.flags);
    }
    MIDBS_freesearch (&sh);

    return;
}

/*
   This function modifies user quota settings.
*/
void set_quota_user (uid_t uid, signed long long Hquota, signed long long Squota, time_t maxT, int set)
{
    search_handle sh;
    struct search_info si;
    struct user_info ui;
    int rc;

    if (MIDBS_allocsearch (&sh) == -1) {
	mylog ("set_quota_user: error: no backend module linked in!");
	exit (EXIT_FAILURE);
    }

    si.uid_first = uid;
    si.uid_last  = uid;
    rc = MIDBS_findfirst (sh, &si, &ui);
    MIDBS_freesearch (&sh);

    if (rc == -2) {
	/* TODO: user does not exist, should do something sensible here */
	/* should not happen */
	mylog ("set_quota_user: warning: user id %lu not yet added to quota database", uid);
	return;
    }

    set_quota_core (&ui, Hquota, Squota, maxT, set);
    MIDBS_write (&ui);

    return;
}

/*
   This function modifies user quota settings.
*/
void set_quota_all (signed long long Hquota, signed long long Squota, time_t maxT, int set)
{
    search_handle sh;
    struct search_info si;
    struct user_info ui;
    int rc;

    if (MIDBS_allocsearch (&sh) == -1) {
	mylog ("set_quota_all: error: no backend module linked in!");
	exit (EXIT_FAILURE);
    }

    si.uid_first = 0;
    si.uid_last  = 1;
    rc = MIDBS_findfirst (sh, &si, &ui);

    if (rc == -2) {
	MIDBS_freesearch (&sh);
	return;
    }

    set_quota_core (&ui, Hquota, Squota, maxT, set);
    MIDBS_write (&ui);

    while (MIDBS_findnext (sh, &ui) == 0) {
	set_quota_core (&ui, Hquota, Squota, maxT, set);
	MIDBS_write (&ui);
    }
    MIDBS_freesearch (&sh);

    return;
}

/*
  Core set quota function
  Input  : ui    : pointer to record with user info
           Hquota: absolute or relative hard quota setting
           Squota: absolute or relative soft quota setting
           maxT  : absolute or relative grace period setting
           set   : indicates relative (0) or absolute (1) setting
  Returns:  0 if everything went ok
           -1 if invalid args
 */
int set_quota_core (struct user_info *ui, signed long long Hquota, signed long long Squota, time_t maxT, int set)
{
    if (set) {
	if (Squota < 0) { 
	    mylog ("set_quota_core: error: negative value for soft quota not allowed (user id %lu)", ui->uid); 
	    return -1;
	}
	if (maxT < 0) { 
	    mylog ("set_quota_core: error: negative value for grace period not allowed (user id %lu)", ui->uid); 
	    return -1;
	}
	ui->Hquota = Hquota;
	ui->Squota = Squota;
	ui->maxT   = maxT;
    } else {
	if ((Squota < 0) && (ui->Squota < (-1 * Squota))) { 
	    mylog ("set_quota_core: error: negative value for soft quota not allowed (user id %lu)", ui->uid); 
	    return -1;
	}
	if ((maxT < 0) && (ui->maxT < (-1 * maxT))) {
	    mylog ("set_quota_core: error: negative value for grace period not allowed (user id %lu)", ui->uid);
	    return -1;
	}
	ui->Hquota += Hquota;
	ui->Squota += Squota;
	ui->maxT   += maxT;
    }
    return 0;
}

/*
  This function sets a user soft quota relative to the hard quota
*/
void set_RSquota_user (uid_t uid, signed long long Squota)
{
    search_handle sh;
    struct search_info si;
    struct user_info ui;
    int rc;

    if (MIDBS_allocsearch (&sh) == -1) {
	mylog ("set_RSquota_user: error: no backend module linked in!");
	exit (EXIT_FAILURE);
    }

    si.uid_first = uid;
    si.uid_last  = uid;
    rc = MIDBS_findfirst (sh, &si, &ui);
    MIDBS_freesearch (&sh);

    if (rc == -2) {
	/* TODO: user does not exist, should do something sensible here */
	/* should not happen */
	mylog ("set_RSquota_user: warning: user id %lu not yet added to quota database", uid);
	return;
    }

    ui.Squota = (ui.Hquota * Squota) / 100;
    MIDBS_write (&ui);

    return;
}

/*
  This function sets a user soft quota relative to the hard quota
*/
void set_RSquota_all (signed long long Squota)
{
    search_handle sh;
    struct search_info si;
    struct user_info ui;
    int rc;

    if (MIDBS_allocsearch (&sh) == -1) {
	mylog ("set_RSquota_all: error: no backend module linked in!");
	exit (EXIT_FAILURE);
    }

    si.uid_first = 0;
    si.uid_last  = 1;
    rc = MIDBS_findfirst (sh, &si, &ui);

    if (rc == -2) {
	MIDBS_freesearch (&sh);
	return;
    }

    ui.Squota = (ui.Hquota * Squota) / 100;
    MIDBS_write (&ui);

    while (MIDBS_findnext (sh, &ui) == 0) {
	ui.Squota = (ui.Hquota * Squota) / 100;
	MIDBS_write (&ui);
    }
    MIDBS_freesearch (&sh);

    return;
}

/* Apply ip consumption to quota */
void modify_quota_user (uid_t uid, unsigned long deltaIP)
{
    search_handle sh;
    struct search_info si;
    struct user_info ui;
    int rc;

    if (MIDBS_allocsearch (&sh) == -1) {
	mylog ("modify_quota_user: error: no backend module linked in!");
	exit (EXIT_FAILURE);
    }

    si.uid_first = uid;
    si.uid_last  = uid;
    rc = MIDBS_findfirst (sh, &si, &ui);
    MIDBS_freesearch (&sh);

    if (rc == -2) {
	/* TODO: user does not exist, should do something sensible here */
	/* should not happen */
	mylog ("modify_quota_user: warning: user id %lu not yet added to quota database", uid);
	return;
    }

    ui.Hquota -= deltaIP;
    MIDBS_write (&ui);

    return;
}

/* Check quota limits of all users */
void check_quota_all (void)
{
    search_handle sh;
    struct search_info si;
    struct user_info ui;
    int rc;

    if (MIDBS_allocsearch (&sh) == -1) {
	mylog ("set_quota_all: error: no backend module linked in!");
	exit (EXIT_FAILURE);
    }

    si.uid_first = 0;
    si.uid_last  = 1;
    rc = MIDBS_findfirst (sh, &si, &ui);

    if (rc == -2) {
	MIDBS_freesearch (&sh);
	return;
    }

    check_quota_core (&ui);
    MIDBS_write (&ui);

    while (MIDBS_findnext (sh, &ui) == 0) {
	check_quota_core (&ui);
	MIDBS_write (&ui);
    }
    MIDBS_freesearch (&sh);

    return;
}

/*
  Core check quota function
  Input  : ui    : pointer to record with user info
  Returns: -
 */
void check_quota_core (struct user_info *ui)
{
    signed long long deltaQ;
    time_t now, deltaT;

    if (ui->Hquota <= 0) {
	/* Hard quota reached: sever connection */
	ui->T0    =  0;
	ui->flags |= IPAU_NO_ACCESS;
	ui->warn  &= ~WARN;
    } else {
	/* Hard quota not yet reached */
	deltaQ = ui->Hquota - ui->Squota;
	if (deltaQ <= 0) {
	    /* Soft quota reached... */
	    if (ui->T0 != 0) {
		/* Soft quota reached earlier */
		time (&now);
		deltaT = now - ui->T0;
		if (deltaT > ui->maxT) {
		    /* Grace period exhausted: sever connection */
		    ui->flags |= IPAU_NO_ACCESS;
		    ui->warn  &= ~WARN;
		} else {
		    /* Still in grace period */
		    ui->flags &= ~IPAU_NO_ACCESS;
		    ui->warn  |= WARN;
		    /* TODO: maybe trigger warning here? */
		}
	    } else {
		/* Soft quota reached just now */
		time (&now);
		ui->T0    =  now;
		ui->flags &= ~IPAU_NO_ACCESS;
		ui->warn  |= WARN;
		/* TODO: trigger warning here! */
	    }
	} else {
	    /* User is in the clear */
	    ui->T0     = 0;
	    ui->flags &= ~IPAU_NO_ACCESS;
	    ui->warn  &= ~WARN;
	}
    }
}

