/*
 * Copyright (C) 1995 Lars Fenneberg
 * Portions Copyright (C) 2000 Erik Cumps
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * $Id: ipadd.c,v 1.17 2000/08/14 19:40:46 slay Exp $
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <signal.h>
#include <getopt.h>
#include <ctype.h>
#include <sys/socket.h>
#include "ipquota.h"
#include "pathnames.h"

char  defaultdb[2048] = _PATH_VAR_LOG_QUOTA_STATS;
char *database;

struct processed_args {
  signed long long   Hquota, Squota;
  time_t             maxT;
  struct passwd     *pwd;
  int                set, do_all, remove;
};

int skfd = -1;

void usage(void)
{
  fprintf (stderr, "Usage: %s [-vh] [-s] <username/uid> <Hquota> <Squota> <grace>\n" \
                   "  or   %s [-vh] -U <username/uid> [-H <Hquota>] [-S <Squota>] [-T <grace>]\n" \
	           "  or   %s [-vh] -R <username/uid>\n\n" \
                 , pname, pname, pname);
  exit (EXIT_FAILURE);
}

void catch_sig(int sig)
{
  close_database (database);
  mylog ("terminating (signal %i)", sig);
  exit (EXIT_FAILURE);
}

/* Parses option argument and returns signed long result */
int arg_to_number (char *arg, signed long long* result, char *msg)
{
  char *lc;
  unsigned long long mul = 1;
  int rel = 0;

  lc = arg+strlen(arg)-1;

  if (!isdigit((*lc))) {
         if ((*lc == 'b') || (*lc == 'B')) mul = 1;
    else if ((*lc == 'k') || (*lc == 'K')) mul = 1024;
    else if ((*lc == 'm') || (*lc == 'M')) mul = 1024*1024;
    else if ((*lc == 'g') || (*lc == 'G')) mul = 1024*1024*1024;
    else if ((*lc == 't') || (*lc == 'T')) mul = (unsigned long long) 1024*1024*1024*1024;
    else if (*lc == '%') { mul = 1; rel = 1; }
    else {
      mylog("malformed %s argument: %s", msg, arg);
      exit (EXIT_FAILURE);
    }
    *lc = '\0';
  }
	
  *result = strtol(arg, &lc, 0) * mul;

  if (*lc!='\0') {
    mylog("malformed %s argument: %s", msg, arg);
    exit (EXIT_FAILURE);
  }			

  return rel;
}


/* Parses option argument and returns time_t result */
int arg_to_time (char *arg, time_t *result, char *msg)
{
  char *lc;
  unsigned long mul = 1;

  lc = arg+strlen(arg)-1;

  if (!isdigit((*lc))) {
         if ((*lc == 's') || (*lc == 'S')) mul = 1;
    else if ((*lc == 'm') || (*lc == 'M')) mul = 60;
    else if ((*lc == 'h') || (*lc == 'H')) mul = 60*60;
    else if ((*lc == 'd') || (*lc == 'D')) mul = 24*60*60;
    else if ((*lc == 'w') || (*lc == 'W')) mul = 7*24*60*60;
    else if ((*lc == 'o') || (*lc == 'O')) mul = 4*7*24*60*60;
    else {
      mylog("malformed %s argument: %s", msg, arg);
      exit (EXIT_FAILURE);
    }
    *lc = '\0';
  }
	
  *result = strtol(arg, &lc, 0) * mul;

  if (*lc!='\0') {
    mylog("malformed %s argument: %s", msg, arg);
    exit (EXIT_FAILURE);
  }

  return 0;
}

/* Parses option argument and returns struct passwd */
int arg_to_uid (char *arg, struct passwd **result, char *msg)
{
  uid_t uid;
  char *c;
  int nan = 0;

  if (*arg == 0) {
      mylog ("arg_to_uid(): empty argument");
      exit (EXIT_FAILURE);
  }

  uid = (uid_t) strtol (arg, &c, 10);
  nan = (*c != 0);

  /* deal with "everyone" special case */
  if (strcmp(arg, "everyone") == 0) {
    return 1;
  }

  /* is it a username? */
  *result = getpwnam (arg);
  if (*result != NULL) {
    uid = (*result)->pw_uid;
  } else {
    /* is it a uid? */
    *result = getpwuid (uid);
    if (nan || (*result == NULL)) {
      /* user not found! */
      mylog ("no such user/uid: %s", arg);
      exit (EXIT_FAILURE);
    }
  }

  /* Refuse to set quota for root */
  if (uid == 0) {
    mylog ("quota not allowed for root");
    exit (EXIT_FAILURE);
  }

  return 0;
}

int process_arguments (int argc, char **argv, struct processed_args *args)
{
  extern int optind;
  extern char *optarg;
  int all3 = 1;
  int Udone = 0, Hdone = 0, Sdone = 0, Srel = 0, Tdone = 0;
  int c;

  args->Hquota = 0;
  args->Squota = 0;
  args->maxT   = 0;
  args->pwd = NULL;
  args->do_all = 0;
  args->set    = 0;
  args->remove = 0;

  while ((c = getopt(argc, argv, "vhsU:H:S:T:R:")) > 0) {
    switch (c) {
    case 'R':
      if (args->set != 0) {
	mylog ("mixing -s and -R options is not allowed");
	exit (EXIT_FAILURE);
      }
      if (!all3) {
	mylog ("mixing -R and -U/-H/-S/-T options not allowed");
	exit (EXIT_FAILURE);
      }
      /* get uid */
      all3 = 0;
      args->remove = 1;
      args->do_all = arg_to_uid (optarg, &args->pwd, "user id");
      if (args->do_all) {
	mylog ("cannot remove user \"everyone\"");
	exit (EXIT_FAILURE);
      }
      break;
    case 'U':
      if (args->set != 0) {
	mylog ("mixing -s and -U options is not allowed");
	exit (EXIT_FAILURE);
      }
      if (args->remove) {
	  mylog ("mixing -R and -U options is not allowed");
	  exit (EXIT_FAILURE);
      }
      /* get uid */
      all3 = 0;
      Udone = 1;
      args->do_all = arg_to_uid (optarg, &args->pwd, "user id");
      break;
    case 'H':
      if (args->set != 0) {
	mylog ("mixing -s and -H options is not allowed");
	exit (EXIT_FAILURE);
      }
      if (args->remove) {
	  mylog ("mixing -R and -H options is not allowed");
	  exit (EXIT_FAILURE);
      }
      /* set Hquota */
      all3 = 0;
      Hdone = 1;
      arg_to_number (optarg, &args->Hquota, "hard quota");
      break;
    case 'S':
      if (args->set != 0) {
	mylog ("mixing -s and -S options is not allowed");
	exit (EXIT_FAILURE);
      }
      if (args->remove) {
	  mylog ("mixing -R and -S options is not allowed");
	  exit (EXIT_FAILURE);
      }
      /* set Squota */
      all3 = 0;
      Sdone = 1;
      Srel = arg_to_number (optarg, &args->Squota, "soft quota");
      break;
    case 'T':
      if (args->set != 0) {
	mylog ("mixing -s and -T options is not allowed");
	exit (EXIT_FAILURE);
      }
      if (args->remove) {
	  mylog ("mixing -R and -T options is not allowed");
	  exit (EXIT_FAILURE);
      }
      /* set maxT */
      arg_to_time (optarg, &args->maxT, "grace period");
      Tdone = 1;
      all3 = 0;
      break;
    case 's':
      if (all3 == 0) {
	mylog ("mixing -s and -U/-H/-S/-T options is not allowed");
	exit (EXIT_FAILURE);
      }
      args->set++;
      break;
    case 'v':
      version();
      break; /* never reached */
    case 'h':
    default:
      usage();
      break; /* never reached */
    }
  }

  if (all3) {
    if ((argc-optind) == 4) {
      /* get uid */
      args->do_all = arg_to_uid (argv[optind], &args->pwd, "user id");

      /* get Hquota */
      arg_to_number (argv[optind+1], &args->Hquota, "hard quota");

      /* get Squota */
      Srel = arg_to_number (argv[optind+2], &args->Squota, "soft quota");
      if (Srel == 1) {
	if (args->set == 0) {
	/* special case: need to modify Squota *after* modifying Hquota */
	  ;
	} else {
	  args->Squota = (args->Hquota * args->Squota) / 100;
	}
      }

      /* get maxT */
      arg_to_time (argv[optind+3], &args->maxT, "grace period");

      if (args->set) {
	if (args->Hquota < 0) {
	  mylog ("negative value for hard quota not allowed");
	  exit (EXIT_FAILURE);
	}
	if (args->Squota < 0) {
	  mylog ("negative value for soft quota not allowed");
	  exit (EXIT_FAILURE);
	}
	if (args->maxT < 0) {
	  mylog ("negative value for grace period not allowed");
	  exit (EXIT_FAILURE);
	}
      }

    } else {
      usage();
    }
  } else {
      if (!args->remove) {
	  if (Udone == 0) {
	      mylog ("-U argument required");
	      exit (EXIT_FAILURE);
	  }
	  if (Srel == 1) {
	      if (Hdone == 0) {
		  mylog ("relative setting of soft quota requires hard quota setting");
		  exit (EXIT_FAILURE);
		  ;
	      } else {
		  args->Squota = (args->Hquota * args->Squota) / 100;
	      }
	  }
	  if (args->Hquota < 0) {
	      mylog ("negative value for hard quota not allowed");
	      exit (EXIT_FAILURE);
	  }
	  if (args->Squota < 0) {
	      mylog ("negative value for soft quota not allowed");
	      exit (EXIT_FAILURE);
	  }
	  if (args->maxT < 0) {
	      mylog ("negative value for grace period not allowed");
	      exit (EXIT_FAILURE);
	  }
	  args->set = 1;
      }
  }

  return Srel;	
}


int main(int argc,char **argv)
{
    struct processed_args args;
    int Srel;

    database = getenv ("IPQUOTADB");
    if (database == NULL) { database = defaultdb; }

    use_syslog = 0;
    pname = argv[0];

    signal(SIGINT, catch_sig);
    signal(SIGTERM, catch_sig);
    signal(SIGQUIT, catch_sig);
    signal(SIGHUP, catch_sig);
	
    /* needs better user interface! */
    Srel = process_arguments (argc, argv, &args);
		
    if ((skfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
	    mylog(my_perror("can't create socket"));
	    exit (EXIT_FAILURE);
	}

    init_hash();
    init_database(database);
	
    if (args.remove) {
	remove_user (args.pwd->pw_uid);
    } else {	
	if (args.do_all) {
	    set_quota_all (args.Hquota, args.Squota, args.maxT, args.set);
	} else {
	    if (exist_user(args.pwd->pw_uid)) {
		set_quota_user (args.pwd->pw_uid, args.Hquota, args.Squota, args.maxT, args.set);
	    } else {
		/* add user with no flags */
		(void) add_user (args.pwd->pw_uid, args.Hquota, args.Squota, args.maxT, 0, 0, 1);
	    }
	}

	/* deal with special case of relative Squota when modifying (!= setting) quota settings*/
	if ((args.set == 0) && (Srel == 1)) {
	    if (args.do_all) {
		set_RSquota_all (args.Squota);
	    } else {
		set_RSquota_user (args.pwd->pw_uid, args.Squota);
	    }
	}


	check_quota_all ();
	ctrl_access_all();
    }

  close_database(database);
  close(skfd);	       

  return EXIT_SUCCESS;
}

