/*
 * Modular IP quota Database Backend System
 * Copyright (C) 2000 Erik Cumps
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * MIDBS classic backend
 * uses classic flatfile text database as in original ipquota
 *
 * $Id: main.c,v 1.26 2000/10/12 20:57:57 slay Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/ip_acct_user.h>
#include <ipquota_shared.h>
#include <midbs.h>
#include <midbs_private.h>

#ifndef lint
static const char _vcid[] = "classic $Id: main.c,v 1.26 2000/10/12 20:57:57 slay Exp $";
#endif /* lint */

/* MIDBS module id */
const char MIDBS_private_id[] = "classic";

/* --- Function prototypes --- */
static int  FetchConnectionArgs      (char *);
static int  read_db                  (void);
static void init_hash                (void);
static int  open_and_lock            (void);
static int  lock_database            (void);
static int  open_database            (void);
static int  fill_hash                (void);
static void unlock_database          (void);
static int  close_db                 (void);
static struct user_hash_s *find_core (uid_t uid);
static int  copy_to_backend          (struct user_info *, struct user_hash_s *);
static int  copy_from_backend        (struct user_hash_s *, struct user_info *);

/* --- hash definitions --- */
#define HASH_TABLE_SIZE		111
#define HASH_FUNC(uid)		((uid) % HASH_TABLE_SIZE)

struct user_hash_s {
    uid_t uid;
    signed long long Hquota, Squota;
    time_t maxT, T0;
    int flags, warn;
    struct user_hash_s *next;	
};

/* --- Search meta-info structure --- */
struct search_meta {
    int in_use;
    uid_t uid_current;
    uid_t uid_last;
};

/* --- other definitions --- */
#define DB_UMASK                0133  /* umask used for database file creation */
#define min(a,b)		(a < b) ? a : b

/* --- Private global variables --- */
static int locked    = 0;
static int updatable = 0;
static int dirty     = 0;
static char *filename;
static char *lockdir;
static FILE *database; 
static struct user_hash_s *hash_table[HASH_TABLE_SIZE];
static uid_t lastuid = 0;

/* --- Exported functions - modular system --- */
int MIDBS_private_open (char *connection)
{
    int rc;

    /* Fetch connection parameters */
    rc = FetchConnectionArgs (connection);
    if (rc != 0) { return -2; }

    /* Read database */
    return read_db ();
}

int MIDBS_private_close (char *connection)
{
    return close_db ();
}

int MIDBS_private_allocsearch (search_handle *handle)
{
    struct search_meta *p;

    if ((p = (struct search_meta *) malloc (sizeof (struct search_meta))) == NULL) {
	mylog (my_perror ("MIDBS_private_allocsearch"));
	return -2;
    }

    p->in_use = 0;
    *handle = (search_handle) p;

    return 0;
}

int MIDBS_private_freesearch (search_handle *handle)
{
    if (*handle == INVALIDSH) {
	return -2;
    }

    free (*handle);
    *handle = INVALIDSH;

    return 0;
}

int MIDBS_private_findfirst (search_handle handle, struct search_info *search, struct user_info *record)
{
    struct search_meta *mysh;
    struct user_hash_s *p = NULL;
    uid_t               uid;

    if (handle == INVALIDSH) {
	return -3;
    }

    /*** FIXME - do more argument checking! */

    mysh = (struct search_meta *) handle;

    if (mysh->in_use == 1) {
        /*** FIXME - Clean up search handle to prepare for new search here */
        /*           just dies for now                                     */
	return -3;
    }

    /* We do only uid searches with this implementation */
    /* Will change someday :)                           */
    if (search->uid_last == 0) {
	mysh->in_use = 0;
	memset ((void *) record, 0, sizeof (struct user_info));
	return -2;
    }

    mysh->in_use = 1;

    if (search->uid_first == 0) {
	/* search all uids */
	mysh->uid_current = 1;
	mysh->uid_last = lastuid;
    } else {
	mysh->uid_current = search->uid_first;
	mysh->uid_last = min (search->uid_last, lastuid);
    }

    /* Look for first existing uid */
    uid = mysh->uid_current;
    while ((uid <= mysh->uid_last)    &&
           (uid >= mysh->uid_current) &&
           ((p = find_core (uid)) == NULL)
	) { uid++; }

    /* Found anything? */
    if (p != NULL) {
	if (mysh->uid_last == uid) {
	    mysh->in_use = 0;
	} else {
	    mysh->uid_current = uid+1;
	}
	copy_from_backend (p, record);
	return 0;
    } else {
	mysh->in_use = 0;
	memset ((void *) record, 0, sizeof (struct user_info));
	return -2;
    }
}

int MIDBS_private_findnext (search_handle handle, struct user_info *record)
{
    struct search_meta *mysh;
    struct user_hash_s *p = NULL;
    uid_t               uid;

    if (handle == INVALIDSH) {
	return -3;
    }

    /*** FIXME - do more argument checking! */

    mysh = (struct search_meta *) handle;

    if (mysh->in_use == 0) {
	return -3;
    }

    /* Look for next existing uid */
    uid = mysh->uid_current;
    while ((uid <= mysh->uid_last)    &&
           (uid >= mysh->uid_current) &&
           ((p = find_core (uid)) == NULL)
	) { uid++; }

    /* Found anything? */
    if (p != NULL) {
	mysh->uid_current = uid+1;
	copy_from_backend (p, record);
	return 0;
    } else {
	mysh->in_use = 0;
	memset ((void *) record, 0, sizeof (struct user_info));
	return -2;
    }
}

int MIDBS_private_write (struct user_info *record)
{
    struct user_hash_s *p;
    int hashval = HASH_FUNC(record->uid);
    int rc;

    p = find_core (record->uid);

    if (p == NULL) {
	if ((p = (struct user_hash_s *) malloc(sizeof (struct user_hash_s))) == NULL) {
	    mylog (my_perror ("MIDBS_private_write"));
	    return -2;
	}
	p->next = hash_table[hashval];
	hash_table[hashval] = p;
	rc = 1;
    } else {
	rc = 0;
    }
    copy_to_backend (record, p);
    if (updatable) { dirty = 1; }

    return rc;
}

int MIDBS_private_delete (struct user_info *record)
{
    struct user_hash_s *p, *q;
        
    for (p = hash_table[HASH_FUNC(record->uid)], q = p; p != NULL; q = p, p = p->next) { 
        if (p->uid == record->uid) break;
    }

    if (p != NULL) {
	if (q == p) {
	    hash_table[HASH_FUNC(record->uid)] = p->next;
	} else {
	    q->next = p->next;
	}
	free (p);
	if (updatable) { dirty = 1; }
	return 0;
    } else {
	return -2;
    }
}


/* --- Private functions --- */

/* Retrieves connection parameters from configuration file
   Input  : connection: name of desired connection
   Returns: 0 if everything went ok
            1 if failed to fetch a required connection parameter
 */
int FetchConnectionArgs (char *connection)
{
    int rc;

    /* Initialise configuration API */
    MIDBS_CFG_Init ((char *)CONFIG_FILE);

    /* Get required connection parameters */
    rc = FetchReqConnectionStr (connection, "file"   , &filename); if (rc != 0) { return 1; }
    rc = FetchReqConnectionStr (connection, "lockdir", &lockdir ); if (rc != 0) { return 1; }

    return 0;
}

/* Opens, locks, reads database
   Input  : -
   Returns:  0 if everything went ok
            -2 if error
 */
int read_db (void)
{
    init_hash ();
    if (open_and_lock () != 0) { return -2; }
    if (fill_hash () != 0) { return -2; }
    return 0;
}

/* Initialise hash
   Input  : -
   Returns: -
*/
void init_hash (void)
{
    int i;
	
    for (i = 0; i < HASH_TABLE_SIZE; i++) hash_table[i] = NULL;
}

/* Open and lock database
   Input  : -
   Returns:  0 if everything went ok
            -2 if error
*/
int open_and_lock (void)
{
    if (lock_database () != 0) { return -2; }
    if (open_database () != 0) { return -2; }
    return 0;
}

/* Lock database
   Input  : -
   Returns:  0 if everything went ok
            -2 if error
 */
int lock_database (void)
{
    char linkfile[1024], lockfile[1024];
    FILE *f;
    int tries = 0;
    int time_to_wait = 1;
    pid_t pid;


    if (locked) {
	mylog ("lock_database: database already locked - ignored");
    } else {
	sprintf (linkfile, "%s/ipacct-lock.%i", lockdir, getpid());
	sprintf (lockfile, "%s/ipacct-lock", lockdir);

	if ((f = fopen (linkfile, "w")) == NULL) {
	    mylog (my_perror (linkfile));
	    return -2;
	}
	fprintf (f, "%i\n", getpid());
	fclose (f);

	while (!locked && (tries < LOCK_TRIES)) {
	    if (link(linkfile, lockfile) != 0) {
		if (errno != EEXIST) {
		    mylog (my_perror ("lockfile hardlink creation"));
		    remove (linkfile);
		    return -2;
		}
		if ((f = fopen (lockfile, "r")) == NULL) {
		    mylog (my_perror (lockfile));
		    remove (linkfile);
		    return 1;
		}
		fscanf (f, "%i\n", (pid_t *) &pid);
		fclose (f);
		mylog ("database file %s locked by process %i", filename, pid);
		if ((kill (pid, 0) != 0) && (errno == ESRCH)) {
		    mylog ("dead process %i - removing old lockfile", pid);
		    if (remove (lockfile) != 0) {
			mylog (my_perror (lockfile));
			remove (linkfile);
			return -2;
		    }
		}
		sleep (time_to_wait);
		time_to_wait *= 2;
		tries++;
	    } else {
		remove (linkfile);
		locked = 1;
	    }
	}

	if (!locked) {
	    mylog ("max tries (%d) exhausted while trying to create lockfile", LOCK_TRIES);
	    remove (linkfile);
	    return -2;
	}
    }
    return 0;
}

/* Open database
   Input  : -
   Returns:  0 if everything went ok
            -2 if error
 */
int open_database (void)
{
    mode_t my_umask;

    if ((database = fopen (filename, "r")) == NULL) {
	if (errno == ENOENT) {
	    my_umask = umask (DB_UMASK);
	    if ((database = fopen (filename, "w")) != NULL) {
		fclose (database);
		umask (my_umask);
		return 0;
	    }
	    umask (my_umask);
	}
	mylog (my_perror (filename));
	unlock_database ();
	return -2;
    } else {
	return 0;
    }
}

/* reads database and fills hash
   Input  : -
   Returns:  0 if everything went ok
            -2 if error
 */
int fill_hash (void)
{
    struct user_info record;
    int line = 0, rc;

#ifdef BENCHMARK
    MIDBS_db_entries = 0;
#endif /* BENCHMARK */
    while (EOF != (rc = fscanf (database, "%i %Li %Li %lu %lu %i %i\n",
			        &record.uid, &record.Hquota, &record.Squota,
                                (unsigned long *) &record.maxT,
                                (unsigned long *) &record.T0, 
                                &record.flags, &record.warn)))
	{
	    line++;
	    if (rc == 7) {
		if (MIDBS_private_write (&record) != 1) {
		    mylog ("fill_hash: problem adding record on line %d in database %s to hash!", line, filename);
		    fclose (database);
		    unlock_database ();
		    return -2;
		} else {
#ifdef BENCHMARK
		    MIDBS_db_entries++;
#endif /* BENCHMARK */
		    if (record.uid > lastuid) { lastuid = record.uid; }
		}
	    } else {
		mylog ("fill_hash: invalid syntax on line %d in database %s!", line, filename);
		fclose (database);
		unlock_database ();
		return -2;
	    }
	}

    fclose (database);
    updatable = 1;
    return 0;
}

/* Unlock database
   Input  : record: holds record data
   Returns:  0 if record was updated
             1 if record was added
            -2 if error
*/
void unlock_database (void)
{
    char temp[1024];
	
    if (locked) {
	sprintf (temp, "%s/ipacct-lock", lockdir);
	remove (temp);
	locked = 0;
    } else {
	mylog ("unlock_database: database not locked - ignored");
    }
}

/* Writes hash to disk and close database
   Input  : -
   Returns:  0 if everything went ok
            -2 if error
 */
int close_db (void)
{
    char temp[1024];
    mode_t my_umask;
    int i;
    struct user_hash_s *p;

    if (dirty) {
	sprintf (temp, "%s.%i.temp", filename, getpid());

	my_umask = umask (DB_UMASK);
	if ((database = fopen (temp, "w")) == NULL) {
	    mylog (my_perror (filename));
	    umask (my_umask);
	    unlock_database ();
	    return -2;
	}

	for(i=0; i<HASH_TABLE_SIZE; i++) {
	    p = hash_table[i];
	    while (p) {
		fprintf (database, "%i %Li %Li %lu %lu %i %i\n", 
			 p->uid, p->Hquota, p->Squota, p->maxT, p->T0, p->flags, p->warn);
		p = p->next;
	    }
	}

	fclose (database);
	umask (my_umask);

	if (rename (temp, filename) != 0) {
	    mylog ("close_db: failed to replace database: %s", my_perror (filename));
	    mylog ("temporary database is %s", temp);
	    unlock_database ();
	    return -2;
	}
    }

    unlock_database ();
    return 0;
}

/* core record search function
   Input  : uid: uid to look for
   Returns: pointer to discovered record
            NULL if not found
 */
struct user_hash_s *find_core (uid_t uid)
{
    struct user_hash_s *p;

#ifdef BENCHMARK
    MIDBS_db_queries++;
#endif /* BENCHMARK */

    for (p = hash_table[HASH_FUNC(uid)]; p != NULL; p = p->next) { 
        if (p->uid == uid) break;
    }

    return p;
}

/* Copies data from client structure to backend structure
   Input  : c: pointer to client structure
            b: pointer to backend structure
   Returns:  0 if everything went ok
            -1 if invalid client pointer
            -2 if invalid backend pointer
 */
int copy_to_backend (struct user_info *c, struct user_hash_s *b)
{
    if (c == NULL) { return -1; }
    if (b == NULL) { return -2; }

    b->uid    = c->uid;
    b->Hquota = c->Hquota;
    b->Squota = c->Squota;
    b->maxT   = c->maxT;
    b->T0     = c->T0;
    b->flags  = c->flags;
    b->warn   = c->warn;

    return 0;
}

/* Copies data from backend structure to client structure
   Input  : b: pointer to backend structure
            c: pointer to client structure
   Returns:  0 if everything went ok
            -1 if invalid client pointer
            -2 if invalid backend pointer
 */
int copy_from_backend (struct user_hash_s *b, struct user_info *c)
{
    if (c == NULL) { return -1; }
    if (b == NULL) { return -2; }

    c->uid    = b->uid;
    c->Hquota = b->Hquota;
    c->Squota = b->Squota;
    c->maxT   = b->maxT;
    c->T0     = b->T0;
    c->flags  = b->flags;
    c->warn   = b->warn;

    return 0;
}

