/*
 * Modular IP quota Database Backend System
 * Copyright (C) 2000 Erik Cumps
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* 
 * MIDBS configuration API
 *
 * $Id: config.c,v 1.8 2000/08/14 19:40:53 slay Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
#include <errno.h>
#include <midbs.h>

#ifndef lint
static const char _vcid[] = "$Id: config.c,v 1.8 2000/08/14 19:40:53 slay Exp $";
#endif /* lint */

/* --- Global variables --- */
static int active = 0;	              /* flag indicating when we're active */
static MIDBS_CFG_section *sectiontop; /* top of linked list of sections    */

/* --- Exported API ---*/

/* Open configuration file and parse contents
   Input:   filename: name of configuration file to read
   Returns: 0 if everything went ok
            1 if failed to open and/or read file

   Note: silently skips errors in configfile > should add verbosity sometime
         should add more checks too
*/
int  MIDBS_CFG_Init (char *filename)
{
    FILE              *cfg;
    MIDBS_CFG_section *sect, *oldsect;
    MIDBS_CFG_entry   *ent, *oldent;
    char              buf[MIDBS_CFG_LINELEN], *p, *q;
    char              flag;

    if (active) {
	return 0;
    } else {
	active = 1;
	/*  Open config file */
	if ((cfg = fopen (filename, "r")) == NULL) { return 1; }

	/*  Extract all valid entries */
	sectiontop = sect = oldsect = NULL;
	ent = oldent = NULL;

	while (!feof(cfg)) {
	    /*  read line */
	    fgets (buf, sizeof(buf), cfg);

            /* discover type: section header or name/value pair? */
	    if (buf[0] == '[') {
                /* get section name */
		p = buf; p++; q = p;
		while ((*q != ']') && *q) { q++; }
		if (*q != ']') { continue; }

                /* create new section entry */
		if ((sect = (MIDBS_CFG_section *) malloc (sizeof(MIDBS_CFG_section))) == NULL) { return 1; }
		sect->entries = NULL;
		sect->next = NULL;
		if (oldsect != NULL) { oldsect->next = sect; }
		oldsect = sect;
		if (sectiontop == NULL) { sectiontop = sect; }

                /* fill out name of section */
		if ((sect->section = (char *) malloc (q-p+1)) == NULL) { return 1; }
		strncpy (sect->section, p, q-p+1);
		sect->section[q-p] = 0x0;

                /* Prepare to hook up name/value pairs */
		oldent = NULL;
	    } else {
                /* is there a section to add to? */
		if (sect == NULL) { continue; }

		/*  isolate entry name and convert to lower case */
		p = buf;
		while (isspace(*p) && *p) { p++; }
		if (!isalpha(*p)) { continue; }
		q = p;
		while (isgraph(*q)) {
		    *q = tolower (*q);
		    q++;
		}

		/*  create new entry */
		if ((ent = (MIDBS_CFG_entry *) malloc (sizeof(MIDBS_CFG_entry))) == NULL) { return 1; }
		ent->next = NULL;
		if (oldent != NULL) { oldent->next = ent; }
		oldent = ent;
		if (sect->entries == NULL) { sect->entries = ent; }

		/*  fill out name of entry */
		if ((ent->name = (unsigned char *) malloc (q-p+1)) == NULL) { return 1; }
		strncpy (ent->name, p, q-p);
		ent->name[q-p] = 0x0;

		/*  scan for entry value */
		p = q;
		while (isspace(*p) && *p) p++;
		if (!(*p)) { ent->value = '\0'; continue; }
		if ((flag = (*p == '"'))) { q = ++p; } else { q = p; }
		while ((flag || !isspace(*q)) && *q) {
		    if (flag && (*q == '"')) {
			if (*(q + 1) == '"') {
			    strcpy (q, q + 1);
			} else {
			    break;
			}
		    }
		    q++;
		}
    
		/*  fill out value of entry */
		if ((ent->value = (unsigned char *) malloc (q-p+1)) == NULL) { return 1; }
		strncpy (ent->value, p, q-p);
		ent->value[q-p] = 0x0;
	    }
	}

	/*  End */
	fclose (cfg);
	return 0;
    }
}

/* Closes configuration file and cleans up memory
   Input  : -
   Returns: -
 */
void MIDBS_CFG_Shutdown (void)
{
    MIDBS_CFG_section *p, *q;
    MIDBS_CFG_entry   *v, *w;

    if (active) {
	active = 0;

        /* Walk through linked list of sections and clean up */
	p = sectiontop;
	while (p != NULL) {
            /* walk */
	    q = p;
	    p = p->next;
            /* clean */
	    v = q->entries;
	    while (v != NULL) {
                /* walk */
		w = v;
		v = v->next;
                /* clean */
		free (w->name);
		free (w->value);
		free (w);
	    }
	    free (q->section);
	    free (q);
	}
    }
}

/* Core GetString function, retrieves value string of section:name
   Input  : section: pointer to name of section
            name   : pointer to name of name
            value  : pointer to storage for pointer to value
   Returns: 0 if value found
            1 if section not found
            2 if name/value not found
*/
int  MIDBS_CFG_GetString_Core (char *section, char *name, char **value)
{
    MIDBS_CFG_section *p;
    MIDBS_CFG_entry   *q;

    *value = NULL;

    p = sectiontop;
    while (p != NULL) {
	if (!strcasecmp (p->section, section)) { break; }
	p = p->next;
    }

    if (p == NULL) {
	return 1;
    }

    q = p->entries;
    while (q != NULL) {
	if (!strcasecmp (q->name, name)) { break; }
	q = q->next;
    }

    if (q == NULL) {
	return 2;
    }

    *value = q->value;
    return 0;
}

/* Dereferences a section name
   Input  : section: pointer to section name
            target : pointer to storage for pointer to name of section target
   Returns: 0 if section succesfully dereferenced
            1 if invalid section or referral target

   Note: if section was not a referral target will contain name of section
 */
int MIDBS_CFG_Deref (char *section, char **target)
{
    char *v, *w;
    int rc;

    w = section;
    rc = MIDBS_CFG_GetString_Core (section, MIDBS_CFG_REFERRAL, &v);
    while (v) {
	w = v;
	rc = MIDBS_CFG_GetString_Core (w, MIDBS_CFG_REFERRAL, &v);
    }
    if (rc == 1) {
	*target = NULL;
	return 1;
    } else {
	*target = w;
	return 0;
    }
}

/* Retrieves value string of section:name, derefs section if necessary
   Input  : section: pointer to name of section
            name   : pointer to name of name
            value  : pointer to storage for pointer to value
   Returns: 0 if value found
            1 if section not found
            2 if name/value not found
 */
int  MIDBS_CFG_GetString (char *section, char *name, char **value)
{
    char *s;

    *value = NULL;
    if (MIDBS_CFG_Deref (section, &s) == 1) {
	return 1;
    }

    return MIDBS_CFG_GetString_Core (s, name, value);
}

/* Retrieves value long of section:name, derefs section if necessary
   Input  : section: pointer to name of section
            name   : pointer to name of name
            value  : pointer to storage for value
   Returns: 0 if value found
            1 if value not found
 */
int  MIDBS_CFG_GetLong (char *section, char *name, long *value)
{
    int   rc;
    char *strvalue;
    char *validity;
    long  result;
  
    rc = MIDBS_CFG_GetString (section, name, &strvalue);
    if (rc) {
	return rc;
    } else {
	result = strtol (strvalue, (char **) &validity, 0);
	/* Check validity */
	if (*validity != '\0') {
	    return 3;
	}
	/* Check value clamping */
	if (errno == ERANGE) {
	    return 3;
	}
	/* ok */
	*value = result;
	return 0;
    }
}

