/*
 * Modular IP quota Database Backend System
 * Copyright (C) 2000 Erik Cumps
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * MIDBS helper code
 *
 * $Id: helper.c,v 1.5 2000/09/01 15:26:34 slay Exp $
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/ip_acct_user.h>
#include <ipquota_shared.h>
#include <midbs.h>
#include <midbs_private.h>

#ifdef BENCHMARK
#include <math.h>
#endif /* BENCHMARK */

#ifndef lint
static const char _vcid[] = "$Id: helper.c,v 1.5 2000/09/01 15:26:34 slay Exp $";
#endif /* lint */

/* --- helper functions --- */

/* controls ip access for uid
   Input  : uid: uid for which to control access
            flags: flags used for access control
   Returns:  0 if everything went ok
            -1 if error
 */
int MIDBS_ctrl_access (uid_t uid, int flags)
{
    struct ip_acct_user_ctrl_entry ct;
    int skfd = -1;
	
    if (uid == 0) return 0;

    ct.uid   = uid;
    ct.flags = flags;

    if ((skfd = socket (AF_INET, SOCK_DGRAM, 0)) < 0) {
	mylog (my_perror ("MIDBS_ctrl_access: can't create socket"));
	return -1;
    }

    if (ioctl (skfd, SIOCCTRLU, &ct) < 0) {
	mylog (my_perror ("MIDBS_ctrl_access: SIOCCTRLU"));
	return -1;
    }

    close (skfd);
    return 0;
}

#ifdef BENCHMARK
/* Reports benchmark results
   Input  : info: pointer to benchmark info
            msg : pointer to benchmark message
   Returns: -
 */
void MIDBS_benchmark_report (struct MIDBS_BM *info, char *msg, unsigned long items, char *item_msg)
{
    clock_t clock_diff;
    double ms_diff;
    double us_diff;

    mylog ("Benchmarking info for %s:", msg);

    if (items) {
	mylog ("For %lu items (%s)", items, item_msg);
    }

    clock_diff = info->clk_stop - info->clk_start;
    mylog ("Clockticks    (wall time)  : %9lu", clock_diff);

    ms_diff = 1e3 * (double) clock_diff / (double) CLK_TCK;
    mylog ("mseconds      (wall time)  : %9.3f", ms_diff);

    if (items) {
	us_diff = 1e6 * (double) clock_diff / (double) CLK_TCK / (double) items;
	mylog ("useconds/item (wall time)  : %9.3f", us_diff);
    }

    clock_diff = info->tms_stop.tms_utime - info->tms_start.tms_utime;
    mylog ("Clockticks    (user time)  : %9lu", clock_diff);

    ms_diff = 1e3 * (double) clock_diff / (double) CLK_TCK;
    mylog ("mseconds      (user time)  : %9.3f", ms_diff);

    if (items) {
	us_diff = 1e6 * (double) clock_diff / (double) CLK_TCK / (double) items;
	mylog ("useconds/item (user time)  : %9.3f", us_diff);
    }

    clock_diff = info->tms_stop.tms_stime - info->tms_start.tms_stime;
    mylog ("Clockticks    (system time): %9lu", clock_diff);

    ms_diff = 1e3 * (double) clock_diff / (double) CLK_TCK;
    mylog ("mseconds      (system time): %9.3f", ms_diff);

    if (items) {
	us_diff = 1e6 * (double) clock_diff / (double) CLK_TCK / (double) items;
	mylog ("useconds/item (system time): %9.3f", us_diff);
    }

    return;
}
#endif /* BENCHMARK */

/* Check if we're suid and drop privileges if so
   Input  : -
   Returns: -
*/
void CheckAndDropSuid (void)
{
    uid_t ruid, euid;
    gid_t rgid, egid;
    static int dropped = 0;

    /* Make sure we drop privileges only once
       calling setreuid() and setregid() twice
       wouldn't do much good
    */
    if (!dropped) {
	ruid = getuid ();
	rgid = getgid ();
	euid = geteuid ();
	egid = getegid ();

	if (euid != ruid) {
	    setreuid (euid, ruid);
	    dropped++;
	    mylog ("CheckAndDropSuid: suid detected: privilege dropped");
	}

	if (egid != rgid) {
	    setregid (egid, rgid);
	    dropped++;
	    mylog ("CheckAndDropSuid: sgid detected: privilege dropped");
	}
    }
}

/* Retrieves required connection string from configuration file
   Input  : connection: name of desired connection
            name      : name of required parameter
            storage   : pointer to storage for result
   Returns: 0 if everything went ok
            1 if failed to fetch string
 */
int FetchReqConnectionStr (char *connection, char *name, char **storage)
{
    int rc;

    rc = MIDBS_CFG_GetString (connection, name, storage);
    if (rc == 1) {
	mylog ("FetchConnectionArgs: Error: Section \"%s\" not found!", connection);
	mylog ("Failed to establish database connection.");
	return 1;
    }
    if (rc == 2) {
	mylog ("FetchConnectionArgs: Error: required \"%s\" key not found in section \"%s\"!", name, connection);
	mylog ("Failed to establish database connection.");
	return 1;
    }
    return 0;
}

/* Retrieves optional connection value from configuration file
   Input  : connection: name of desired connection
            name      : name of required parameter
            storage   : pointer to storage for result
   Returns: -
 */
int FetchConnectionVal (char *connection, char *name, long *storage)
{
    int rc;

    rc = MIDBS_CFG_GetLong (connection, name, storage);
    if (rc == 3) {
	mylog ("FetchConnectionArgs: Warning: invalid value found for \"%s\" key in section \"%s\".", name, connection);
	mylog ("FetchConnectionArgs: Using default value (%d)", *storage);
    }
    return 0;
}
