/*
 * Modular IP quota Database Backend System
 * Copyright (C) 2000 Erik Cumps
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * MIDBS main code
 *
 * $Id: main.c,v 1.19 2000/09/01 15:27:47 slay Exp $
 */

#include <stdio.h>
#include <ipquota_shared.h>
#include <midbs.h>
#include <midbs_private.h>
#include <dlfcn.h>

#ifdef BENCHMARK
#include <time.h>
#include <sys/times.h>
#endif /* BENCHMARK */

#ifndef lint
static const char _vcid[] = "$Id: main.c,v 1.19 2000/09/01 15:27:47 slay Exp $";
#endif /* lint */

/* --- global variables --- */
static int linked = 0;

#ifdef BENCHMARK
unsigned long MIDBS_db_entries = 0;
unsigned long MIDBS_db_queries = 0;
struct MIDBS_BM runningtime;
#endif /* BENCHMARK */

/* --- API pointer structure --- */
static struct MIDBS_API {
    char *id;
    int (       *open)(char *);
    int (      *close)(char *);
    int (*allocsearch)(search_handle *);
    int ( *freesearch)(search_handle *);
    int (  *findfirst)(search_handle, struct search_info *, struct user_info *);
    int (   *findnext)(search_handle, struct user_info *);
    int (      *write)(struct user_info *);
    int (     *delete)(struct user_info *);
    void *dyn_handle;
} API = {
    "unlinked",
    &MIDBS_open_stub,
    &MIDBS_close_stub,
    &MIDBS_allocsearch_stub,
    &MIDBS_freesearch_stub, 
    &MIDBS_findfirst_stub,
    &MIDBS_findnext_stub,
    &MIDBS_write_stub,
    &MIDBS_delete_stub
};

/* --- symbol checker --- */
int MIDBS_sym_check (void *symbol, char *symbolname)
{
    if (symbol == NULL) {
	mylog ("MIDBS_sym_check: symbol %s not found", symbolname);
	dlclose (API.dyn_handle);
	return 1;
    }
    return 0;
}

/* --- MIDBS backend initialiser --- */
int MIDBS_open (char *connection)
{
    char *p;
    int rc;

#ifdef BENCHMARK
    struct MIDBS_BM linking, backend;
#endif /* BENCHMARK */

    if (linked) {
	mylog ("MIDBS_open: already linked with a database backend module!");
	return -1;
    } else {

#ifdef BENCHMARK
	linking.clk_start = times(&linking.tms_start);
#endif /* BENCHMARK */

	/* Initialise configuration API */
	MIDBS_CFG_Init ((char *)CONFIG_FILE);

        /* Check and drop suid/sgid privileges if necessary */
	CheckAndDropSuid ();

	/* Get filename of desired module */
	rc = MIDBS_CFG_GetString (connection, "module", &p);
	if (rc == 1) {
	    mylog ("No section found for connection \"%s\"! Failed to link in backend module.\n", connection);
	    return -1;
	}
	if (rc == 2) {
	    mylog ("No module key found in section \"%s\"! Failed to link in backend module.\n", connection);
	    return -1;
	}

	/* Link in desired module */
	if ((API.dyn_handle = dlopen (p, RTLD_NOW)) == NULL) {
	    mylog ("MIDBS_open: dlopen: %s", dlerror());
	    return -1;
	}

	/* Update API pointers */
	API.id          = dlsym (API.dyn_handle, "MIDBS_private_id"         );
	if (MIDBS_sym_check (API.id,          "MIDBS_private_id"         )) { return -1; }
	API.open        = dlsym (API.dyn_handle, "MIDBS_private_open"       );
	if (MIDBS_sym_check (API.open,        "MIDBS_private_open"       )) { return -1; }
	API.close       = dlsym (API.dyn_handle, "MIDBS_private_close"      );
	if (MIDBS_sym_check (API.close,       "MIDBS_private_close"      )) { return -1; }
	API.allocsearch = dlsym (API.dyn_handle, "MIDBS_private_allocsearch");
	if (MIDBS_sym_check (API.allocsearch, "MIDBS_private_allocsearch")) { return -1; }
	API.freesearch  = dlsym (API.dyn_handle, "MIDBS_private_freesearch" );
	if (MIDBS_sym_check (API.freesearch,  "MIDBS_private_freesearch" )) { return -1; }
	API.findfirst   = dlsym (API.dyn_handle, "MIDBS_private_findfirst"  );
	if (MIDBS_sym_check (API.findfirst,   "MIDBS_private_findfirst"  )) { return -1; }
	API.findnext    = dlsym (API.dyn_handle, "MIDBS_private_findnext"   );
	if (MIDBS_sym_check (API.findnext,    "MIDBS_private_findnext"   )) { return -1; }
	API.write       = dlsym (API.dyn_handle, "MIDBS_private_write"      );
	if (MIDBS_sym_check (API.write,       "MIDBS_private_write"      )) { return -1; }
	API.delete      = dlsym (API.dyn_handle, "MIDBS_private_delete"     );
	if (MIDBS_sym_check (API.delete,      "MIDBS_private_delete"     )) { return -1; }
	linked = 1;

#ifdef BENCHMARK
	linking.clk_stop = times(&linking.tms_stop);
#endif /* BENCHMARK */

        /* Report module id */
	mylog ("%s backend module linked in.", API.id);

#ifdef BENCHMARK
	backend.clk_start = times(&backend.tms_start);
#endif /* BENCHMARK */

	/* Jump to module's open function */
	rc = API.open (connection);

#ifdef BENCHMARK
	backend.clk_stop = times(&backend.tms_stop);
	MIDBS_benchmark_report (&linking, "[midbs] backend module link-in", 0, "");
	MIDBS_benchmark_report (&backend, "[backend] database connection", MIDBS_db_entries, "database entries");
#endif /* BENCHMARK */

	if (rc) {
	    mylog ("Failed to create connection \"%s\" with database", connection);
	}

#ifdef BENCHMARK
	runningtime.clk_start = times(&runningtime.tms_start);
#endif /* BENCHMARK */

	return rc;
    }
}

/* --- API handlers --- */

int MIDBS_close (char *connection)
{
    int rc;

#ifdef BENCHMARK
    struct MIDBS_BM backend;
    runningtime.clk_stop = times(&runningtime.tms_stop);
    backend.clk_start = times(&backend.tms_start);
#endif /* BENCHMARK */

    rc = API.close (connection);
    if (dlclose (API.dyn_handle) != 0) {
	mylog ("MIDBS_close: dlclose: %s", dlerror());
    }

#ifdef BENCHMARK
    backend.clk_stop = times(&backend.tms_stop);
    MIDBS_benchmark_report (&runningtime, "[process] running time",  MIDBS_db_queries, "database queries");
    MIDBS_benchmark_report (&backend, "[backend] database shutdown", MIDBS_db_entries, "database entries");
#endif /* BENCHMARK */

	return rc;
}

int MIDBS_allocsearch (search_handle *handle)
{
    return API.allocsearch (handle);
}

int MIDBS_freesearch (search_handle *handle)
{
    return API.freesearch (handle);
}

int MIDBS_findfirst (search_handle handle, struct search_info *search, struct user_info *record)
{
    return API.findfirst (handle, search, record);
}

int MIDBS_findnext (search_handle handle, struct user_info *record)
{
    return API.findnext (handle, record);
}

int MIDBS_write (struct user_info *record)
{
    return API.write (record);
}

int MIDBS_delete (struct user_info *record)
{
    return API.delete (record);
}

/* --- MIDBS API stub functions --- */

int MIDBS_open_stub (char *connection)
{
    return -1;
}

int MIDBS_close_stub (char *connection)
{
    return -1;
}

int MIDBS_allocsearch_stub (search_handle *handle)
{
    return -1;
}

int MIDBS_freesearch_stub (search_handle *handle)
{
    return -1;
}

int MIDBS_findfirst_stub (search_handle handle, struct search_info *search, struct user_info *record)
{
    return -1;
}

int MIDBS_findnext_stub (search_handle handle, struct user_info *record)
{
    return -1;
}

int MIDBS_write_stub (struct user_info *record)
{
    return -1;
}

int MIDBS_delete_stub (struct user_info *record)
{
    return -1;
}

/* --- weak MIDBS backend functions --- */

/* We use a gcc extension and declare the set of backend MIDBS_private* functions
   here. When a backend module is linked in the linker will replace these symbols
   with the new symbols (because they are weak). This prevents errors when a backend
   module 'forgets' to export some MIDBS_private* functions and the MIDBS main
   code tries to obtain pointers to them after linking in the module.
*/

int MIDBS_private_close       (char *) __attribute__ ((weak));
int MIDBS_private_allocsearch (search_handle *) __attribute__ ((weak));
int MIDBS_private_freesearch  (search_handle *) __attribute__ ((weak));
int MIDBS_private_findfirst   (search_handle, struct search_info *, struct user_info *) __attribute__ ((weak));
int MIDBS_private_findnext    (search_handle, struct user_info *) __attribute__ ((weak));
int MIDBS_private_write       (struct user_info *) __attribute__ ((weak));
int MIDBS_private_delete      (struct user_info *) __attribute__ ((weak));

int MIDBS_private_close (char *connection)
{
    mylog ("Warning: linked in backend module does not provide MIDBS_private_close!");
    return -1;
}

int MIDBS_private_allocsearch (search_handle *handle)
{
    mylog ("Warning: linked in backend module does not provide MIDBS_private_allocsearch!");
    return -1;
}

int MIDBS_private_freesearch  (search_handle *handle)
{
    mylog ("Warning: linked in backend module does not provide MIDBS_private_freesearch!");
    return -1;
}


int MIDBS_private_findfirst (search_handle handle, struct search_info *search, struct user_info *record)
{
    mylog ("Warning: linked in backend module does not provide MIDBS_private_findfirst!");
    return -1;
}

int MIDBS_private_findnext (search_handle handle, struct user_info *record)
{
    mylog ("Warning: linked in backend module does not provide MIDBS_private_findnext!");
    return -1;
}

int MIDBS_private_write (struct user_info *record)
{
    mylog ("Warning: linked in backend module does not provide MIDBS_private_write!");
    return -1;
}

int MIDBS_private_delete (struct user_info *record)
{
    mylog ("Warning: linked in backend module does not provide MIDBS_private_delete!");
    return -1;
}

