/*
 * Modular IP quota Database Backend System
 * Copyright (C) 2000 Erik Cumps
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* 
 * MIDBS MySQL backend
 * uses MySQL database to store ip quota information
 *
 * $Id: main.c,v 1.12 2000/10/12 20:57:59 slay Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <linux/ip_acct_user.h>
#include <mysql/mysql.h>
#include <ipquota_shared.h>
#include <midbs.h>
#include <midbs_private.h>
#include "midbs_mysql.h"

#ifndef lint
static const char _vcid[] = "MySQL $Id: main.c,v 1.12 2000/10/12 20:57:59 slay Exp $";
#endif /* lint */

/* MIDBS module id */
const char MIDBS_private_id[] = "MySQL";

/* --- Function prototypes --- */
static struct user_hash_s *find_core	(uid_t);
static int  copy_to_backend     	(struct user_info *, struct user_hash_s *);
static int  copy_from_backend   	(struct user_hash_s *, struct user_info *);
static int  FetchConnectionArgs		(char *);
static int  GetLock                     (void);
static int  ReleaseLock                 (void);
static int  GetMaxuid                   (void);
static int  get_uid                     (uid_t);
static long query_core                  (char *, char *, MYSQL *, MYSQL_RES **, MYSQL_ROW **);
static void init_hash                   (void);
static struct user_hash_s *find_uid     (uid_t);

#ifdef BENCHMARK
static int  GetDBEntries                (void);
#endif /* BENCHMARK */

/* --- hash definitions --- */
#define HASH_TABLE_SIZE		111
#define HASH_FUNC(uid)		((uid) % HASH_TABLE_SIZE)

/* --- internal user info structure --- */
struct user_hash_s {
    uid_t uid;
    signed long long Hquota, Squota;
    time_t maxT, T0;
    int flags, warn;
    struct user_hash_s *next;	
};

/* --- Search meta-info structure --- */
struct search_meta {
    int in_use;
    uid_t uid_current;
    uid_t uid_last;
};

/* --- other definitions --- */
#define min(a,b)	(a < b) ? a : b
#define max(a,b)	(a > b) ? a : b

/* --- Private global variables --- */
static MYSQL mysql;

/* required config parameters */
static char *host     = NULL;
static char *user     = NULL;
static char *password = NULL;
static char *database = NULL;
static char *table    = NULL;

/* optional config parameters */
static long  port     = 0;
static long  UCD      = 100;
static long  FullUC   = 1;

static struct user_hash_s *hash_table[HASH_TABLE_SIZE];
static uid_t firstuid = 1;
/* silly kludge for now, must implement better search system */
// static uid_t lastuid  = 65535;
static uid_t lastuid  = 0;
static uid_t max_uid  = 0;
static uid_t min_uid  = 1;

static char *lockname    = "IPQUOTA_LOCK";
static long  locktimeout = 10;
static long  locktries   = 3;
static int   locked      = 0;

/* --- Exported functions - modular system --- */
int MIDBS_private_open (char *connection)
{
    int rc;

    /* Fetch connection parameters */
    rc = FetchConnectionArgs (connection);
    if (rc != 0) { return -2; }

    /* Report a few config settings */
    mylog ("Caching: %s", FullUC ? "full record caching" : "uid caching only");
    mylog ("Caching: readahead: %lu %s", UCD, FullUC ? "records" : "uids");

    /* Create new MySQL object */
    if (!mysql_init (&mysql)) {
	mylog ("open: MySQL reported insufficient memory while creating MYSQL object.");
	return -2;
    }

    /* Set up connection with MySQL server */
    mysql_real_connect (&mysql, host, user, password, database, port, NULL, 0);
    CHECK_SQL_ERROR_R("open", "failed to connect to database", &mysql, -2);

    /* Try to obtain a lock (cooperative advisory locking) */
    rc = GetLock ();
    if (rc != 0) {
	MIDBS_private_close (connection);
	return -2;
    }

#ifdef BENCHMARK
    /* Discover number of database entries */
    rc = GetDBEntries ();
    if (rc != 0) {
	MIDBS_private_close (connection);
	return -2;
    }
#endif /* BENCHMARK */

    /* Discover largest uid */
    rc = GetMaxuid ();
    if (rc != 0) {
	MIDBS_private_close (connection);
	return -2;
    }

    return 0;
}

int MIDBS_private_close (char *connection)
{
    int rc;

    /* release lock (cooperative advisory locking) */
    rc = ReleaseLock ();
    if (rc) { mylog ("close: ignoring failure to release lock"); }

    /* Close connection with MySQL server */
    mysql_close (&mysql);
    if (mysql_errno(&mysql)) {
	mylog ("close: failed to close connection with database: error: %s (%u)", mysql_error (&mysql), mysql_errno (&mysql));
	return -2;
    }

    return 0;
}

int MIDBS_private_allocsearch (search_handle *handle)
{
    struct search_meta *p;

    if ((p = (struct search_meta *) malloc (sizeof (struct search_meta))) == NULL) {
	mylog (my_perror ("MIDBS_private_allocsearch"));
	return -2;
    }

    p->in_use = 0;
    *handle = (search_handle) p;

    return 0;
}

int MIDBS_private_freesearch (search_handle *handle)
{
    if (*handle == INVALIDSH) {
	return -2;
    }

    free (*handle);
    *handle = INVALIDSH;

    return 0;
}

int MIDBS_private_findfirst (search_handle handle, struct search_info *search, struct user_info *record)
{
    struct search_meta *mysh;
    struct user_hash_s *p = NULL;
    uid_t               uid;

    if (handle == INVALIDSH) {
	return -3;
    }

    /*** FIXME - do more argument checking! */

    mysh = (struct search_meta *) handle;

    if (mysh->in_use == 1) {
        /*** FIXME - Clean up search handle to prepare for new search here */
        /*           just dies for now                                     */
	return -3;
    }

    /* We do only uid searches with this implementation */
    /* Will change someday :)                           */
    if (search->uid_last == 0) {
	mysh->in_use = 0;
	memset ((void *) record, 0, sizeof (struct user_info));
	return -2;
    }

    mysh->in_use = 1;

    if (search->uid_first == 0) {
	/* search all uids */
	mysh->uid_current = 1;
	mysh->uid_last = lastuid;
    } else {
	mysh->uid_current = search->uid_first;
	mysh->uid_last = min (search->uid_last, lastuid);
    }

    /* Look for first existing uid */
    uid = mysh->uid_current;
    while ((uid <= mysh->uid_last)    &&
           (uid >= mysh->uid_current) &&
           ((p = find_uid (uid)) == NULL)
	) { uid++; }

    /* Found anything? */
    if (p != NULL) {
	if (mysh->uid_last == uid) {
	    mysh->in_use = 0;
	} else {
	    mysh->uid_current = uid+1;
	}
        /*** FIXME - optimize this away */
	if (!FullUC) {
	    p = find_core (uid);
	}
	copy_from_backend (p, record);
        /* Whoops! Serious bug here, was removing cached records without updating hash structure */
	if (!FullUC) {
	    free (p);
	}
	return 0;
    } else {
	mysh->in_use = 0;
	memset ((void *) record, 0, sizeof (struct user_info));
	return -2;
    }
}

int MIDBS_private_findnext (search_handle handle, struct user_info *record)
{
    struct search_meta *mysh;
    struct user_hash_s *p = NULL;
    uid_t               uid;

    if (handle == INVALIDSH) {
	return -3;
    }

    /*** FIXME - do more argument checking! */

    mysh = (struct search_meta *) handle;

    if (mysh->in_use == 0) {
	return -3;
    }

    /* Look for next existing uid */
    uid = mysh->uid_current;
    while ((uid <= mysh->uid_last)    &&
           (uid >= mysh->uid_current) &&
           ((p = find_uid (uid)) == NULL)
	) { uid++; }

    /* Found anything? */
    if (p != NULL) {
	mysh->uid_current = uid+1;
        /*** FIXME - optimize this away */
	if (!FullUC) {
	    p = find_core (uid);
	}
	copy_from_backend (p, record);
        /* Whoops! Serious bug here, was removing cached records without updating hash structure */
	if (!FullUC) {
	    free (p);
	}
	return 0;
    } else {
	mysh->in_use = 0;
	memset ((void *) record, 0, sizeof (struct user_info));
	return -2;
    }
}

int MIDBS_private_write (struct user_info *record)
{
    char                query[256];
    MYSQL_RES          *result;
    MYSQL_ROW          *rows;
    long                rc;
    struct user_hash_s *p;

    sprintf (query, "replace into %s values(%i, %Li, %Li, %lu, %lu, %i, %i)",
             table, record->uid, record->Hquota, record->Squota, record->maxT,
             record->T0, record->flags, record->warn);
    rc = query_core ("MIDBS_private_write", query, &mysql, &result, &rows);

    if (rc < 0) {
	mylog ("MIDBS_private_write: query failed");
	return -2;
    }

    if (rc > 0) {
	mylog ("MIDBS_private_write: weird: query returned rows");
	mylog ("MIDBS_private_write: don't trust this, return weird error");
	/* query returned data: cleanup is ok */
	mysql_free_result (result);
	free (rows);
	return -1;
    }

    rc = mysql_affected_rows (&mysql);
    if (rc == 0) {
	mylog ("MIDBS_private_write: weird: replace claims no affected rows");
	mylog ("MIDBS_private_write: don't trust this, return weird error");
	rc = -1;
    } else {
	if (rc > 2) {
	    mylog ("MIDBS_private_write: weird: query affected more than one record");
	    mylog ("MIDBS_private_write: don't trust this, return weird error");
	    rc = -1;
	} else {
	    if (rc == 2) { rc = 0; }
            /* update cache */
	    p = find_uid (record->uid);
	    if (p == NULL) {
		if ((p = (struct user_hash_s*) calloc (1, sizeof (struct user_hash_s))) == NULL) {
		    mylog ("MIDBS_private_write: out of memory updating cache");
		    mylog ("MIDBS_private_write: this is fatal, return weird error");
                    /*** FIXME - deal with this in a better way */
		    return -1;
		}
		copy_to_backend (record, p);
		p->next = hash_table[HASH_FUNC(p->uid)];
		hash_table[HASH_FUNC(p->uid)] = p;
	    } else {
	        copy_to_backend (record, p);
	    }
	}
    }

    return rc;

}

int MIDBS_private_delete (struct user_info *record)
{
    char                query[256];
    MYSQL_RES          *result;
    MYSQL_ROW          *rows;
    long                rc;
    struct user_hash_s *p, *q;

    sprintf (query, "delete from %s where uid=%i", table, record->uid);
    rc = query_core ("MIDBS_private_delete", query, &mysql, &result, &rows);

    if (rc < 0) {
	mylog ("MIDBS_private_delete: query failed");
	return -2;
    }

    if (rc > 0) {
	mylog ("MIDBS_private_delete: weird: query returned rows");
	mylog ("MIDBS_private_delete: don't trust this, return weird error");
	/* query returned data: cleanup is ok */
	mysql_free_result (result);
	free (rows);
	return -1;
    }

    rc = mysql_affected_rows (&mysql);
    if (rc == 0) {
	mylog ("MIDBS_private_delete: uid not found");
	rc = -2;
    } else {
	if (rc > 1) {
	    mylog ("MIDBS_private_delete: weird: query deleted more than one record");
	    mylog ("MIDBS_private_delete: don't trust this, return weird error");
	    rc = -1;
	} else {
	    rc = 0;
            /* update cache */
	    for (p = hash_table[HASH_FUNC(record->uid)], q = p; p != NULL; q = p, p = p->next) { 
		if (p->uid == record->uid) break;
	    }
	    if (p != NULL) {
		if (q == p) {
		    hash_table[HASH_FUNC(record->uid)] = p->next;
		} else {
		    q->next = p->next;
		}
		free (p);
		return 0;
	    }
	}
    }

    return rc;
}


/* --- Private functions --- */

/* core record search function
   Input  : uid: uid to look for
   Returns: pointer to discovered record
            NULL if not found
 */
struct user_hash_s *find_core (uid_t uid)
{
    struct user_hash_s *p;
    char query[256];
    MYSQL_RES *result;
    MYSQL_ROW *rows;
    long rc;

    /* no need to do all this work if we cache records */
    if (FullUC) {
	return find_uid (uid);
    }

    p = (struct user_hash_s *)NULL;

    sprintf (query, "select Hquota, Squota, maxT, T0, flags, warn from %s where uid=%i", table, uid);
    rc = query_core ("find_core", query, &mysql, &result, &rows);

    if (rc < 0) {
	mylog ("find_core: query failed");
	return p;
    }

    if (rc > 1) {
	mylog ("find_core: weird: got more than one row for uid %lu", uid);
	mylog ("find_core: weird: results ignored!");
	/* query returned data: cleanup is ok */
	mysql_free_result (result);
	free (rows);
	return p;
    }

    if (rc == 0) {
	return p;
    }

    p = (struct user_hash_s *) malloc (sizeof(struct user_hash_s));
    if (p == NULL) {
	mylog ("find_core: error: out of memory allocating result storage: %s (%d)", strerror(errno), errno);
	/* query returned data: cleanup is ok */
	mysql_free_result (result);
	free (rows);
	return p;
    }

    /* no checking of sscanf result, mysql checks values for us at entry */
    p->uid = uid;
    sscanf (rows[0][0], "%Li", &p->Hquota);
    sscanf (rows[0][1], "%Li", &p->Squota);
    sscanf (rows[0][2], "%lu", &p->maxT  );
    sscanf (rows[0][3], "%lu", &p->T0    );
    sscanf (rows[0][4], "%i" , &p->flags );
    sscanf (rows[0][5], "%i" , &p->warn  );

    /* query returned data: cleanup is ok */
    mysql_free_result (result);
    free (rows);
    return p;
}

/* Copies data from client structure to backend structure
   Input  : c: pointer to client structure
            b: pointer to backend structure
   Returns:  0 if everything went ok
            -1 if invalid client pointer
            -2 if invalid backend pointer
 */
int copy_to_backend (struct user_info *c, struct user_hash_s *b)
{
    if (c == NULL) { return -1; }
    if (b == NULL) { return -2; }

    b->uid    = c->uid;
    b->Hquota = c->Hquota;
    b->Squota = c->Squota;
    b->maxT   = c->maxT;
    b->T0     = c->T0;
    b->flags  = c->flags;
    b->warn   = c->warn;

    return 0;
}

/* Copies data from backend structure to client structure
   Input  : b: pointer to backend structure
            c: pointer to client structure
   Returns:  0 if everything went ok
            -1 if invalid client pointer
            -2 if invalid backend pointer
 */
int copy_from_backend (struct user_hash_s *b, struct user_info *c)
{
    if (c == NULL) { return -1; }
    if (b == NULL) { return -2; }

    c->uid    = b->uid;
    c->Hquota = b->Hquota;
    c->Squota = b->Squota;
    c->maxT   = b->maxT;
    c->T0     = b->T0;
    c->flags  = b->flags;
    c->warn   = b->warn;

    return 0;
}


/* Retrieves connection parameters from configuration file
   Input  : connection: name of desired connection
   Returns: 0 if everything went ok
            1 if failed to fetch a required connection parameter
 */
int FetchConnectionArgs (char *connection)
{
    int rc;

    /* Initialise configuration API */
    MIDBS_CFG_Init ((char *)CONFIG_FILE);

    /* Get required connection parameters */
    rc = FetchReqConnectionStr (connection, "host"    , &host    ); if (rc != 0) { return 1; }
    rc = FetchReqConnectionStr (connection, "user"    , &user    ); if (rc != 0) { return 1; }
    rc = FetchReqConnectionStr (connection, "password", &password); if (rc != 0) { return 1; }
    rc = FetchReqConnectionStr (connection, "database", &database); if (rc != 0) { return 1; }
    rc = FetchReqConnectionStr (connection, "table",    &table   ); if (rc != 0) { return 1; }

    /* Get optional connection parameters */
    FetchConnectionVal (connection, "port",   &port);
    FetchConnectionVal (connection, "UCD",    &UCD);
    FetchConnectionVal (connection, "FullUC", &FullUC);

    return 0;
}

/* Tries to obtain a lock for cooperative advisory locking
   Input  : -
   Returns: 0 if everything went ok
            1 if failed to get lock
 */
int GetLock (void)
{
    char       query[128];
    int        lock = 0;
    MYSQL_RES *result;
    MYSQL_ROW *rows;
    long       rc;

    if (!locked) {
	sprintf (query, "select GET_LOCK(\"%s\",%lu)", lockname, locktimeout);

	while (!locked && (lock <= locktries)) {
	    lock++;
	    rc = query_core ("GetLock", query, &mysql, &result, &rows);

	    if (rc < 0) { 
		mylog ("GetLock: query failed");
		return 1;
	    }

            if (rc == 0) {
		mylog ("GetLock: query returned no results");
		return 1;
	    }

	    if (rows[0][0] == NULL) {
		REPORT_QUERY_ERROR_MSG ("GetLock", query, "server error");
	    } else {
		locked = atoi (rows[0][0]);
	    }

            /* query returned data: cleanup is ok */
	    mysql_free_result (result);
	    free (rows);
	}

	if (!locked) {
	    mylog ("GetLock: failed to get lock after %lu tries", lock);
	    return 1;
	} else {
	    mylog ("GetLock: got lock after %lu tries", lock);
	    return 0;
	}

    } else {
	return 0;
    }
}

/* Releases the lock obtained for cooperative advisory locking
   Input  : -
   Returns: 0 if everything went ok
            1 if failed to release lock
 */
int ReleaseLock (void)
{
    char       query[128];
    MYSQL_RES *result;
    MYSQL_ROW *rows;
    long       rc;

    if (locked) {
	sprintf (query, "select RELEASE_LOCK(\"%s\")", lockname);
        rc = query_core ("ReleaseLock", query, &mysql, &result, &rows);

	if (rc < 0) {
	    mylog ("ReleaseLock: query failed");
	    return 1;
	}

	if (rc == 0) {
	    mylog ("ReleaseLock: query returned no results");
	    return 1;
	}

	rc = 1;
	if (rows[0][0] == NULL) {
	    mylog ("ReleaseLock: lock \"%s\" does not exist", lockname);
	} else {
	    locked -= atoi (rows[0][0]);
	    if (locked) {
		mylog ("ReleaseLock: lock \"%s\" was not obtained by this client", lockname);
	    } else {
		mylog ("ReleaseLock: lock \"%s\" released", lockname);
		rc = 0;
	    }
	}

        /* query returned data: cleanup is ok */
	mysql_free_result (result);
	free (rows);
	
	return rc;
    } else {
	return 0;
    }
}

#ifdef BENCHMARK
/* Discovers number of database entries
   Input  : -
   Returns: 0 if everything went ok
            1 if SQL query failure
*/
int GetDBEntries (void)
{
    char       query[128];
    MYSQL_RES *result;
    MYSQL_ROW *rows;
    long       rc;

    MIDBS_db_entries = 0;

    sprintf (query, "select count(*) from %s", table);
    rc = query_core ("GetDBEntries", query, &mysql, &result, &rows);

    if (rc < 0) {
	mylog ("GetDBEntries: query failed");
	return 1;
    }

    if (rc == 0) {
	mylog ("GetDBEntries: query returned no results");
	return 1;
    }

    MIDBS_db_entries = atoi (rows[0][0]);

    /* query returned data: cleanup is ok */
    mysql_free_result (result);
    free (rows);

    return 0;
}
#endif /* BENCHMARK */

/* Discovers largest uid
   Input  : -
   Returns: 0 if everything went ok
            1 if SQL query failure
*/
int GetMaxuid (void)
{
    char       query[128];
    MYSQL_RES *result;
    MYSQL_ROW *rows;
    long       rc;

    sprintf (query, "select max(uid) from %s", table);
    rc = query_core ("GetMaxuid", query, &mysql, &result, &rows);

    if (rc < 0) {
	mylog ("GetMaxuid: query failed");
	return 1;
    }

    if (rc == 0) {
	mylog ("GetMaxuid: query returned no results");
	return 1;
    }

    /* check for NULL result */
    if (rows[0][0] != NULL) {
	lastuid = atoi (rows[0][0]);
    } else {
	lastuid = 0;
    }

    /* query returned data: cleanup is ok */
    mysql_free_result (result);
    free (rows);

    return 0;
}

/* lookup function for uid, or record, caching
   Input  : uid_t
   Returns: 0 if everything went ok
            1 if SQL query failure
            2 if out of memory
*/
int get_uid (uid_t uid)
{
    char                query[128];
    uid_t               start, stop;
    MYSQL_RES          *result;
    MYSQL_ROW          *rows;
    long                rc, i;
    struct user_hash_s *p;

    /* already cached ? */
    if ((uid >= min_uid) && (uid <= max_uid)) { return 0; }

    /* discover what range of uids to get */
    if (!max_uid && min_uid) {
	init_hash ();
	start = uid;
	stop  = uid + UCD;
    } else {
	if (uid > max_uid) {
	    start = max_uid + 1;
	    stop  = max (uid, max_uid + UCD);
	} else {
	    start = min (uid, min_uid - UCD - 1);
	    stop  = min_uid - 1;
	}
    }

    /* take absolute boundaries into account */
    start = max (start, firstuid);
    stop  = min (stop,  lastuid);

    /* update boundary info of cached region */
    min_uid = min (start, min_uid);
    max_uid = max (stop , max_uid);

    /* decide what query to run */
    sprintf (query, "select uid%s from %s where uid>=%i and uid<=%i",
             FullUC ? ", Hquota, Squota, maxT, T0, flags, warn" : "",
             table, start, stop);
    rc = query_core ("get_uid", query, &mysql, &result, &rows);

    if (rc < 0) {
	mylog ("get_uid: query failed");
	return 1;
    }

    if (rc == 0) {
	mylog ("get_uid: query returned no results");
	return 1;
    }

    for (i = 0; i < rc; i++) {
	if ((p = (struct user_hash_s *) calloc (1, sizeof(struct user_hash_s))) == NULL) {
	    mylog ("get_uid: out of memory allocating room for uid hash");
	    return 2;
	}

	p->uid = atol (rows[i][0]);
        /* check if we need to process more */
	if (FullUC) {
            /* no checking of sscanf result, mysql checks values for us at entry */
	    sscanf (rows[i][1], "%Li", &p->Hquota);
	    sscanf (rows[i][2], "%Li", &p->Squota);
	    sscanf (rows[i][3], "%lu", &p->maxT  );
	    sscanf (rows[i][4], "%lu", &p->T0    );
	    sscanf (rows[i][5], "%i" , &p->flags );
	    sscanf (rows[i][6], "%i" , &p->warn  );
	}

        p->next = hash_table[HASH_FUNC(p->uid)];
	hash_table[HASH_FUNC(p->uid)] = p;
    }

    mysql_free_result (result);
    free (rows);

    return 0;
}

/* Performs a MySQL SQL query and returns results in array of MYSQL_ROW's
   As MYSQL_ROW is basically a char ** that points into an MYSQL_RES structure
   it seemed pointless to alloc new mem for that and copy the data. So instead
   we just create an array of pointers into the MySQL client lib memspace.

   !!! The caller is responsible for cleanup by freeing the array of rows 
       and calling mysql_free_result !!!

   Input  : func  : string with name of calling function
            query : string with query to execute
            obj   : pointer to mysql object
            result: pointer to storage for pointer to MySQL result
                    NULL if failure
            rows  : pointer to storage for pointer to array of result rows
                    NULL if failure
   Returns: number of result rows (>=0) if everything went ok
            -1     if SQL query failure
            -2     if other failure
*/
long query_core (char *func, char *query, MYSQL *obj, MYSQL_RES **result, MYSQL_ROW **rows)
{
    long       nr, i;

#ifdef BENCHMARK
    MIDBS_db_queries++;
#endif /* BENCHMARK */

    /* preset rows */
    *rows = NULL;

    /* exec query */
    mysql_query (obj, query);
    CHECK_QUERY_ERROR_R (func, query, obj, -1);

    /* store results in client */
    *result = mysql_store_result (obj);
    CHECK_QUERY_ERROR_R(func, query, obj, -1);

    /* check for empty result set */
    if (*result == NULL) {
	return 0;
    }

    /* discover number of rows in result */
    nr = (long) mysql_num_rows (*result);
    if (nr == 0) {
        /* I believe a NULL pointer is returned for empty result sets -> no free needed */
	*result = NULL;
	return 0;
    }

    /* allocate memory for the array of rows */
    *rows = (MYSQL_ROW *) calloc (nr, sizeof(MYSQL_ROW));
    if (*rows == NULL) {
	mylog ("query_core: error allocating storage for %lu rows: %s (%d)", nr, strerror(errno), errno);
	mysql_free_result (*result);
	*result = NULL;
	*rows   = NULL;
	return -2;
    }

    /* fetch and store the rows */
    for (i=0; i<nr; i++) {
	(*rows)[i] = mysql_fetch_row (*result);
	if ((*rows)[i] == NULL) {
	    REPORT_QUERY_ERROR_MSG (func, query, "unexpected end of rows");
	    mysql_free_result (*result);
	    free (*rows);
	    *result = NULL;
	    *rows   = NULL;
	    return -1;
	}
    }

    /* return results */
    return nr;
}

/* --- Hash related functions --- */

/* Initialise hash
   Input  : -
   Returns: -
*/
void init_hash (void)
{
    int i;
	
    for (i = 0; i < HASH_TABLE_SIZE; i++) hash_table[i] = NULL;
}

/* hash search function
   Input  : uid: uid to look for
   Returns: pointer to discovered record
            NULL if not found
 */
struct user_hash_s *find_uid (uid_t uid)
{
    struct user_hash_s *p;

    if (get_uid (uid) != 0) {
	mylog ("find_uid: error caching uid info!");
	return NULL;
    }

    for (p = hash_table[HASH_FUNC(uid)]; p != NULL; p = p->next) { 
        if (p->uid == uid) break;
    }

    return p;
}

